@@ -40,6 +40,8 @@
 
 /* --- prototypes --- */
 static void gtk_accel_group_finalize	(GObject		*object);
+static void accel_closure_invalidate    (gpointer  data,
+                                         GClosure *closure);
 
 
 /* --- variables --- */
@@ -54,7 +56,7 @@ static guint		 default_accel_mod_mask = (GDK_SHIFT_MASK |
 						   GDK_META_MASK);
 
 
-G_DEFINE_TYPE (GtkAccelGroup, gtk_accel_group, G_TYPE_OBJECT);
+G_DEFINE_TYPE (GtkAccelGroup, gtk_accel_group, G_TYPE_OBJECT)
 
 /* --- functions --- */
 static void
@@ -130,10 +132,11 @@ gtk_accel_group_finalize (GObject *object)
 
       if (entry->accel_path_quark)
 	{
-	  const gchar *accel_path = g_quark_to_string (entry[i].accel_path_quark);
+	  const gchar *accel_path = g_quark_to_string (entry->accel_path_quark);
 
 	  _gtk_accel_map_remove_group (accel_path, accel_group);
 	}
+      g_closure_remove_invalidate_notifier (entry->closure, accel_group, accel_closure_invalidate);
     }
 
   g_free (accel_group->priv_accels);
@@ -394,10 +397,10 @@ quick_accel_add (GtkAccelGroup  *accel_group,
 
 static void
 quick_accel_remove (GtkAccelGroup      *accel_group,
-		    GtkAccelGroupEntry *entry)
+                    guint               pos)
 {
-  guint pos = entry - accel_group->priv_accels;
   GQuark accel_quark = 0;
+  GtkAccelGroupEntry *entry = accel_group->priv_accels + pos;
   guint accel_key = entry->key.accel_key;
   GdkModifierType accel_mods = entry->key.accel_mods;
   GClosure *closure = entry->closure;
@@ -574,7 +577,7 @@ gtk_accel_group_disconnect (GtkAccelGroup *accel_group,
     if (accel_group->priv_accels[i].closure == closure)
       {
 	g_object_ref (accel_group);
-	quick_accel_remove (accel_group, accel_group->priv_accels + i);
+	quick_accel_remove (accel_group, i);
 	g_object_unref (accel_group);
 	return TRUE;
       }
