@@ -46,6 +46,7 @@
 #include "gtktoolbutton.h"
 #include "gtktoolbar.h"
 #include "gtkprivate.h"
+#include "gtkbuildable.h"
 #include "gtkalias.h"
 
 
@@ -53,7 +54,7 @@
 
 struct _GtkActionPrivate 
 {
-  gchar *name;
+  const gchar *name; /* interned */
   gchar *label;
   gchar *short_label;
   gchar *tooltip;
@@ -107,14 +108,22 @@ enum
   PROP_ACTION_GROUP
 };
 
+/* GtkBuildable */
+static void gtk_action_buildable_init             (GtkBuildableIface *iface);
+static void gtk_action_buildable_set_name         (GtkBuildable *buildable,
+						   const gchar  *name);
+static const gchar* gtk_action_buildable_get_name (GtkBuildable *buildable);
+
+G_DEFINE_TYPE_WITH_CODE (GtkAction, gtk_action, G_TYPE_OBJECT,
+			 G_IMPLEMENT_INTERFACE (GTK_TYPE_BUILDABLE,
+						gtk_action_buildable_init))
+
 
 static GQuark      accel_path_id  = 0;
 static GQuark      quark_gtk_action_proxy  = 0;
 static const gchar accel_path_key[] = "GtkAction::accel_path";
 static const gchar gtk_action_proxy_key[] = "gtk-action";
 
-G_DEFINE_TYPE (GtkAction, gtk_action, G_TYPE_OBJECT)
-
 static void gtk_action_finalize     (GObject *object);
 static void gtk_action_set_property (GObject         *object,
 				     guint            prop_id,
@@ -378,12 +387,37 @@ gtk_action_init (GtkAction *action)
   action->private_data->proxies = NULL;
 }
 
+static void
+gtk_action_buildable_init (GtkBuildableIface *iface)
+{
+  iface->set_name = gtk_action_buildable_set_name;
+  iface->get_name = gtk_action_buildable_get_name;
+}
+
+static void
+gtk_action_buildable_set_name (GtkBuildable *buildable,
+			       const gchar  *name)
+{
+  GtkAction *action = GTK_ACTION (buildable);
+
+  action->private_data->name = g_intern_string (name);
+}
+
+static const gchar *
+gtk_action_buildable_get_name (GtkBuildable *buildable)
+{
+  GtkAction *action = GTK_ACTION (buildable);
+
+  return action->private_data->name;
+}
+
 /**
  * gtk_action_new:
  * @name: A unique name for the action
- * @label: the label displayed in menu items and on buttons
- * @tooltip: a tooltip for the action
- * @stock_id: the stock icon to display in widgets representing the action
+ * @label: the label displayed in menu items and on buttons, or %NULL
+ * @tooltip: a tooltip for the action, or %NULL
+ * @stock_id: the stock icon to display in widgets representing the
+ *   action, or %NULL
  *
  * Creates a new #GtkAction object. To add the action to a
  * #GtkActionGroup and set the accelerator for the action,
@@ -401,16 +435,14 @@ gtk_action_new (const gchar *name,
 		const gchar *tooltip,
 		const gchar *stock_id)
 {
-  GtkAction *action;
+  g_return_val_if_fail (name != NULL, NULL);
 
-  action = g_object_new (GTK_TYPE_ACTION,
-			 "name", name,
-			 "label", label,
-			 "tooltip", tooltip,
-			 "stock-id", stock_id,
-			 NULL);
-
-  return action;
+  return g_object_new (GTK_TYPE_ACTION,
+                       "name", name,
+		       "label", label,
+		       "tooltip", tooltip,
+		       "stock-id", stock_id,
+		       NULL);
 }
 
 static void
@@ -419,7 +451,6 @@ gtk_action_finalize (GObject *object)
   GtkAction *action;
   action = GTK_ACTION (object);
 
-  g_free (action->private_data->name);
   g_free (action->private_data->label);
   g_free (action->private_data->short_label);
   g_free (action->private_data->tooltip);
@@ -440,16 +471,13 @@ gtk_action_set_property (GObject         *object,
 			 GParamSpec      *pspec)
 {
   GtkAction *action;
-  gchar *tmp;
   
   action = GTK_ACTION (object);
 
   switch (prop_id)
     {
     case PROP_NAME:
-      tmp = action->private_data->name;
-      action->private_data->name = g_value_dup_string (value);
-      g_free (tmp);
+      action->private_data->name = g_intern_string (g_value_get_string (value));
       break;
     case PROP_LABEL:
       gtk_action_set_label (action, g_value_get_string (value));
@@ -509,7 +537,7 @@ gtk_action_get_property (GObject    *object,
   switch (prop_id)
     {
     case PROP_NAME:
-      g_value_set_string (value, action->private_data->name);
+      g_value_set_static_string (value, action->private_data->name);
       break;
     case PROP_LABEL:
       g_value_set_string (value, action->private_data->label);
@@ -609,7 +637,8 @@ _gtk_action_sync_menu_visible (GtkAction *action,
 			       GtkWidget *proxy,
 			       gboolean   empty)
 {
-  gboolean visible, hide_if_empty;
+  gboolean visible = TRUE;
+  gboolean hide_if_empty = TRUE;
 
   g_return_if_fail (GTK_IS_MENU_ITEM (proxy));
   g_return_if_fail (action == NULL || GTK_IS_ACTION (action));
@@ -617,8 +646,12 @@ _gtk_action_sync_menu_visible (GtkAction *action,
   if (action == NULL)
     action = g_object_get_qdata (G_OBJECT (proxy), quark_gtk_action_proxy);
 
-  visible = gtk_action_is_visible (action);
-  hide_if_empty = action->private_data->hide_if_empty;
+  if (action)
+    {
+      /* a GtkMenu for a <popup/> doesn't have to have an action */
+      visible = gtk_action_is_visible (action);
+      hide_if_empty = action->private_data->hide_if_empty;
+    }
 
   if (visible && !(empty && hide_if_empty))
     gtk_widget_show (proxy);
@@ -1415,15 +1448,8 @@ static void
 gtk_action_sync_tooltip (GtkAction *action,
 			 GtkWidget *proxy)
 {
-  GtkWidget *parent;
-
-  parent = gtk_widget_get_parent (proxy);
-  
-  if (GTK_IS_TOOL_ITEM (proxy) && GTK_IS_TOOLBAR (parent))
-    gtk_tool_item_set_tooltip (GTK_TOOL_ITEM (proxy), 
-			       GTK_TOOLBAR (parent)->tooltips,
-			       action->private_data->tooltip,
-			       NULL);
+  gtk_tool_item_set_tooltip_text (GTK_TOOL_ITEM (proxy),
+				  action->private_data->tooltip);
 }
 
 static void 
@@ -1441,8 +1467,9 @@ gtk_action_set_tooltip (GtkAction   *action,
   for (p = action->private_data->proxies; p; p = p->next)
     {
       proxy = (GtkWidget *)p->data;
-      
-      gtk_action_sync_tooltip (action, proxy);
+
+      if (GTK_IS_TOOL_ITEM (proxy))
+        gtk_action_sync_tooltip (action, proxy);
     }
 
   g_object_notify (G_OBJECT (action), "tooltip");
@@ -1617,8 +1644,6 @@ static void
 gtk_action_set_action_group (GtkAction	    *action,
 			     GtkActionGroup *action_group)
 {
-  g_return_if_fail (GTK_IS_ACTION (action));
-
   if (action->private_data->action_group == NULL)
     g_return_if_fail (GTK_IS_ACTION_GROUP (action_group));
   else
