@@ -80,9 +80,12 @@ G_DEFINE_TYPE (GtkPathBar, gtk_path_bar, GTK_TYPE_CONTAINER)
 
 static void gtk_path_bar_finalize                 (GObject          *object);
 static void gtk_path_bar_dispose                  (GObject          *object);
+static void gtk_path_bar_realize                  (GtkWidget        *widget);
+static void gtk_path_bar_unrealize                (GtkWidget        *widget);
 static void gtk_path_bar_size_request             (GtkWidget        *widget,
 						   GtkRequisition   *requisition);
-static void gtk_path_bar_unmap                    (GtkWidget          *widget);
+static void gtk_path_bar_map                      (GtkWidget        *widget);
+static void gtk_path_bar_unmap                    (GtkWidget        *widget);
 static void gtk_path_bar_size_allocate            (GtkWidget        *widget,
 						   GtkAllocation    *allocation);
 static void gtk_path_bar_add                      (GtkContainer     *container,
@@ -93,11 +96,17 @@ static void gtk_path_bar_forall                   (GtkContainer     *container,
 						   gboolean          include_internals,
 						   GtkCallback       callback,
 						   gpointer          callback_data);
-static void gtk_path_bar_scroll_up                (GtkWidget        *button,
-						   GtkPathBar       *path_bar);
-static void gtk_path_bar_scroll_down              (GtkWidget        *button,
-						   GtkPathBar       *path_bar);
+static gboolean gtk_path_bar_scroll               (GtkWidget        *widget,
+						   GdkEventScroll   *event);
+static void gtk_path_bar_scroll_up                (GtkPathBar       *path_bar);
+static void gtk_path_bar_scroll_down              (GtkPathBar       *path_bar);
 static void gtk_path_bar_stop_scrolling           (GtkPathBar       *path_bar);
+static gboolean gtk_path_bar_slider_up_defocus    (GtkWidget        *widget,
+						   GdkEventButton   *event,
+						   GtkPathBar       *path_bar);
+static gboolean gtk_path_bar_slider_down_defocus  (GtkWidget        *widget,
+						   GdkEventButton   *event,
+						   GtkPathBar       *path_bar);
 static gboolean gtk_path_bar_slider_button_press  (GtkWidget        *widget,
 						   GdkEventButton   *event,
 						   GtkPathBar       *path_bar);
@@ -132,10 +141,17 @@ get_slider_button (GtkPathBar  *path_bar,
 		   GtkArrowType arrow_type)
 {
   GtkWidget *button;
+  AtkObject *atk_obj;
 
   gtk_widget_push_composite_child ();
 
   button = gtk_button_new ();
+  atk_obj = gtk_widget_get_accessible (button);
+  if (arrow_type == GTK_ARROW_LEFT)
+    atk_object_set_name (atk_obj, _("Up Path"));
+  else
+    atk_object_set_name (atk_obj, _("Down Path"));
+
   gtk_button_set_focus_on_click (GTK_BUTTON (button), FALSE);
   gtk_container_add (GTK_CONTAINER (button), gtk_arrow_new (arrow_type, GTK_SHADOW_OUT));
   gtk_container_add (GTK_CONTAINER (path_bar), button);
@@ -162,8 +178,11 @@ gtk_path_bar_init (GtkPathBar *path_bar)
   path_bar->down_slider_button = get_slider_button (path_bar, GTK_ARROW_RIGHT);
   path_bar->icon_size = FALLBACK_ICON_SIZE;
   
-  g_signal_connect (path_bar->up_slider_button, "clicked", G_CALLBACK (gtk_path_bar_scroll_up), path_bar);
-  g_signal_connect (path_bar->down_slider_button, "clicked", G_CALLBACK (gtk_path_bar_scroll_down), path_bar);
+  g_signal_connect_swapped (path_bar->up_slider_button, "clicked", G_CALLBACK (gtk_path_bar_scroll_up), path_bar);
+  g_signal_connect_swapped (path_bar->down_slider_button, "clicked", G_CALLBACK (gtk_path_bar_scroll_down), path_bar);
+
+  g_signal_connect (path_bar->up_slider_button, "focus_out_event", G_CALLBACK (gtk_path_bar_slider_up_defocus), path_bar);
+  g_signal_connect (path_bar->down_slider_button, "focus_out_event", G_CALLBACK (gtk_path_bar_slider_down_defocus), path_bar);
 
   g_signal_connect (path_bar->up_slider_button, "button_press_event", G_CALLBACK (gtk_path_bar_slider_button_press), path_bar);
   g_signal_connect (path_bar->up_slider_button, "button_release_event", G_CALLBACK (gtk_path_bar_slider_button_release), path_bar);
@@ -188,12 +207,16 @@ gtk_path_bar_class_init (GtkPathBarClass *path_bar_class)
   gobject_class->dispose = gtk_path_bar_dispose;
 
   widget_class->size_request = gtk_path_bar_size_request;
+  widget_class->realize = gtk_path_bar_realize;
+  widget_class->unrealize = gtk_path_bar_unrealize;
+  widget_class->map = gtk_path_bar_map;
   widget_class->unmap = gtk_path_bar_unmap;
   widget_class->size_allocate = gtk_path_bar_size_allocate;
   widget_class->style_set = gtk_path_bar_style_set;
   widget_class->screen_changed = gtk_path_bar_screen_changed;
   widget_class->grab_notify = gtk_path_bar_grab_notify;
   widget_class->state_changed = gtk_path_bar_state_changed;
+  widget_class->scroll_event = gtk_path_bar_scroll;
 
   container_class->add = gtk_path_bar_add;
   container_class->forall = gtk_path_bar_forall;
@@ -264,7 +287,6 @@ remove_settings_signal (GtkPathBar *path_bar,
 static void
 gtk_path_bar_dispose (GObject *object)
 {
-  GList *list;
   GtkPathBar *path_bar = GTK_PATH_BAR (object);
 
   remove_settings_signal (path_bar, gtk_widget_get_screen (GTK_WIDGET (object)));
@@ -347,14 +369,67 @@ gtk_path_bar_update_slider_buttons (GtkPathBar *path_bar)
     }
 }
 
+static void
+gtk_path_bar_map (GtkWidget *widget)
+{
+  gdk_window_show (GTK_PATH_BAR (widget)->event_window);
+
+  GTK_WIDGET_CLASS (gtk_path_bar_parent_class)->map (widget);
+}
+
 static void
 gtk_path_bar_unmap (GtkWidget *widget)
 {
   gtk_path_bar_stop_scrolling (GTK_PATH_BAR (widget));
+  gdk_window_hide (GTK_PATH_BAR (widget)->event_window);
 
   GTK_WIDGET_CLASS (gtk_path_bar_parent_class)->unmap (widget);
 }
 
+static void
+gtk_path_bar_realize (GtkWidget *widget)
+{
+  GtkPathBar *path_bar;
+  GdkWindowAttr attributes;
+  gint attributes_mask;
+
+  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
+
+  path_bar = GTK_PATH_BAR (widget);
+  widget->window = gtk_widget_get_parent_window (widget);
+  g_object_ref (widget->window);
+
+  attributes.window_type = GDK_WINDOW_CHILD;
+  attributes.x = widget->allocation.x;
+  attributes.y = widget->allocation.y;
+  attributes.width = widget->allocation.width;
+  attributes.height = widget->allocation.height;
+  attributes.wclass = GDK_INPUT_ONLY;
+  attributes.event_mask = gtk_widget_get_events (widget);
+  attributes.event_mask |= GDK_SCROLL_MASK;
+  attributes_mask = GDK_WA_X | GDK_WA_Y;
+
+  path_bar->event_window = gdk_window_new (gtk_widget_get_parent_window (widget),
+							&attributes, attributes_mask);
+  gdk_window_set_user_data (path_bar->event_window, widget);
+
+  widget->style = gtk_style_attach (widget->style, widget->window);
+}
+
+static void
+gtk_path_bar_unrealize (GtkWidget *widget)
+{
+  GtkPathBar *path_bar;
+
+  path_bar = GTK_PATH_BAR (widget);
+
+  gdk_window_set_user_data (path_bar->event_window, NULL);
+  gdk_window_destroy (path_bar->event_window);
+  path_bar->event_window = NULL;
+
+  (* GTK_WIDGET_CLASS (gtk_path_bar_parent_class)->unrealize) (widget);
+}
+
 /* This is a tad complicated
  */
 static void
@@ -375,6 +450,11 @@ gtk_path_bar_size_allocate (GtkWidget     *widget,
 
   widget->allocation = *allocation;
 
+  if (GTK_WIDGET_REALIZED (widget))
+    gdk_window_move_resize (path_bar->event_window,
+			    allocation->x, allocation->y,
+			    allocation->width, allocation->height);
+
   /* No path is set; we don't have to allocate anything. */
   if (path_bar->button_list == NULL)
     return;
@@ -579,6 +659,25 @@ gtk_path_bar_screen_changed (GtkWidget *widget,
   gtk_path_bar_check_icon_theme (GTK_PATH_BAR (widget));
 }
 
+static gboolean
+gtk_path_bar_scroll (GtkWidget      *widget,
+		     GdkEventScroll *event)
+{
+  switch (event->direction)
+    {
+    case GDK_SCROLL_RIGHT:
+    case GDK_SCROLL_DOWN:
+      gtk_path_bar_scroll_down (GTK_PATH_BAR (widget));
+      break;
+    case GDK_SCROLL_LEFT:
+    case GDK_SCROLL_UP:
+      gtk_path_bar_scroll_up (GTK_PATH_BAR (widget));
+      break;
+    }
+
+  return TRUE;
+}
+
 static void
 gtk_path_bar_add (GtkContainer *container,
 		  GtkWidget    *widget)
@@ -664,7 +763,7 @@ gtk_path_bar_forall (GtkContainer *container,
 }
 
 static void
-gtk_path_bar_scroll_down (GtkWidget *button, GtkPathBar *path_bar)
+gtk_path_bar_scroll_down (GtkPathBar *path_bar)
 {
   GList *list;
   GList *down_button = NULL;
@@ -680,11 +779,17 @@ gtk_path_bar_scroll_down (GtkWidget *button, GtkPathBar *path_bar)
       return;   
     }
 
+  if (gtk_widget_get_child_visible (BUTTON_DATA (path_bar->button_list->data)->button))
+    {
+      /* Return if the last button is already visible */
+      return;
+    }
+
   gtk_widget_queue_resize (GTK_WIDGET (path_bar));
 
   border_width = GTK_CONTAINER (path_bar)->border_width;
   direction = gtk_widget_get_direction (GTK_WIDGET (path_bar));
-  
+
   /* We find the button at the 'down' end that we have to make
    * visible */
   for (list = path_bar->button_list; list; list = list->next)
@@ -707,7 +812,10 @@ gtk_path_bar_scroll_down (GtkWidget *button, GtkPathBar *path_bar)
 	}
     }
 
-  space_needed = BUTTON_DATA (down_button->data)->button->allocation.width + path_bar->spacing;
+  /* We check if down_button might be NULL in cases where the pathbar's horizontal size is smaller
+   * than the button and it doesn't get displayed. e.g., on Windows it might be "My Documents and Settings"
+   */
+  space_needed = down_button ? BUTTON_DATA (down_button->data)->button->allocation.width : 0 + path_bar->spacing;
   if (direction == GTK_TEXT_DIR_RTL)
     space_available = path_bar->down_slider_button->allocation.x - GTK_WIDGET (path_bar)->allocation.x;
   else
@@ -727,7 +835,7 @@ gtk_path_bar_scroll_down (GtkWidget *button, GtkPathBar *path_bar)
 }
 
 static void
-gtk_path_bar_scroll_up (GtkWidget *button, GtkPathBar *path_bar)
+gtk_path_bar_scroll_up (GtkPathBar *path_bar)
 {
   GList *list;
 
@@ -737,9 +845,17 @@ gtk_path_bar_scroll_up (GtkWidget *button, GtkPathBar *path_bar)
       return;   
     }
 
+  list = g_list_last (path_bar->button_list);
+
+  if (gtk_widget_get_child_visible (BUTTON_DATA (list->data)->button))
+    {
+      /* Return if the first button is already visible */
+      return;
+    }
+
   gtk_widget_queue_resize (GTK_WIDGET (path_bar));
 
-  for (list = g_list_last (path_bar->button_list); list; list = list->prev)
+  for ( ; list; list = list->prev)
     {
       if (list->prev && gtk_widget_get_child_visible (BUTTON_DATA (list->prev->data)->button))
 	{
@@ -756,14 +872,12 @@ gtk_path_bar_scroll_timeout (GtkPathBar *path_bar)
 {
   gboolean retval = FALSE;
 
-  GDK_THREADS_ENTER ();
-
   if (path_bar->timer)
     {
       if (path_bar->scrolling_up)
-	gtk_path_bar_scroll_up (path_bar->up_slider_button, path_bar);
+	gtk_path_bar_scroll_up (path_bar);
       else if (path_bar->scrolling_down)
-	gtk_path_bar_scroll_down (path_bar->down_slider_button, path_bar);
+	gtk_path_bar_scroll_down (path_bar);
 
       if (path_bar->need_timer) 
 	{
@@ -774,7 +888,7 @@ gtk_path_bar_scroll_timeout (GtkPathBar *path_bar)
 
 	  path_bar->need_timer = FALSE;
 
-	  path_bar->timer = g_timeout_add (timeout * SCROLL_DELAY_FACTOR,
+	  path_bar->timer = gdk_threads_add_timeout (timeout * SCROLL_DELAY_FACTOR,
 					   (GSourceFunc)gtk_path_bar_scroll_timeout,
 					   path_bar);
 	}
@@ -782,8 +896,6 @@ gtk_path_bar_scroll_timeout (GtkPathBar *path_bar)
 	retval = TRUE;
     }
 
-  GDK_THREADS_LEAVE ();
-
   return retval;
 }
 
@@ -798,6 +910,62 @@ gtk_path_bar_stop_scrolling (GtkPathBar *path_bar)
     }
 }
 
+static gboolean
+gtk_path_bar_slider_up_defocus (GtkWidget      *widget,
+                                    GdkEventButton *event,
+                                    GtkPathBar     *path_bar)
+{
+  GList *list;
+  GList *up_button = NULL;
+
+  if (event->type != GDK_FOCUS_CHANGE)
+    return FALSE;
+
+  for (list = g_list_last (path_bar->button_list); list; list = list->prev)
+    {
+      if (gtk_widget_get_child_visible (BUTTON_DATA (list->data)->button))
+        {
+          up_button = list;
+          break;
+        }
+    }
+
+  /* don't let the focus vanish */
+  if ((!GTK_WIDGET_IS_SENSITIVE (path_bar->up_slider_button)) || 
+      (!gtk_widget_get_child_visible (path_bar->up_slider_button)))
+    gtk_widget_grab_focus (BUTTON_DATA (up_button->data)->button);
+
+  return FALSE;
+}
+
+static gboolean
+gtk_path_bar_slider_down_defocus (GtkWidget      *widget,
+                                    GdkEventButton *event,
+                                    GtkPathBar     *path_bar)
+{
+  GList *list;
+  GList *down_button = NULL;
+
+  if (event->type != GDK_FOCUS_CHANGE)
+    return FALSE;
+
+  for (list = path_bar->button_list; list; list = list->next)
+    {
+      if (gtk_widget_get_child_visible (BUTTON_DATA (list->data)->button))
+        {
+          down_button = list;
+          break;
+        }
+    }
+
+  /* don't let the focus vanish */
+  if ((!GTK_WIDGET_IS_SENSITIVE (path_bar->down_slider_button)) || 
+      (!gtk_widget_get_child_visible (path_bar->down_slider_button)))
+    gtk_widget_grab_focus (BUTTON_DATA (down_button->data)->button);
+
+  return FALSE;
+}
+
 static gboolean
 gtk_path_bar_slider_button_press (GtkWidget      *widget, 
 				  GdkEventButton *event,
@@ -812,13 +980,13 @@ gtk_path_bar_slider_button_press (GtkWidget      *widget,
     {
       path_bar->scrolling_down = FALSE;
       path_bar->scrolling_up = TRUE;
-      gtk_path_bar_scroll_up (path_bar->up_slider_button, path_bar);
+      gtk_path_bar_scroll_up (path_bar);
     }
   else if (widget == path_bar->down_slider_button)
     {
       path_bar->scrolling_up = FALSE;
       path_bar->scrolling_down = TRUE;
-      gtk_path_bar_scroll_down (path_bar->down_slider_button, path_bar);
+      gtk_path_bar_scroll_down (path_bar);
     }
 
   if (!path_bar->timer)
@@ -829,7 +997,7 @@ gtk_path_bar_slider_button_press (GtkWidget      *widget,
       g_object_get (settings, "gtk-timeout-initial", &timeout, NULL);
 
       path_bar->need_timer = TRUE;
-      path_bar->timer = g_timeout_add (timeout,
+      path_bar->timer = gdk_threads_add_timeout (timeout,
 				       (GSourceFunc)gtk_path_bar_scroll_timeout,
 				       path_bar);
     }
@@ -1149,8 +1317,7 @@ button_data_free (ButtonData *button_data)
     gtk_file_path_free (button_data->path);
   button_data->path = NULL;
 
-  if (button_data->dir_name)
-    g_free (button_data->dir_name);
+  g_free (button_data->dir_name);
   button_data->dir_name = NULL;
 
   button_data->button = NULL;
@@ -1256,22 +1423,16 @@ button_drag_data_get_cb (GtkWidget          *widget,
 {
   ButtonData *button_data;
   GtkPathBar *path_bar;
-  char *uri;
-  char *uri_list;
+  char *uris[2];
 
   button_data = data;
   path_bar = GTK_PATH_BAR (widget->parent); /* the button's parent *is* the path bar */
 
-  uri = gtk_file_system_path_to_uri (path_bar->file_system, button_data->path);
-  uri_list = g_strconcat (uri, "\r\n", NULL);
-  g_free (uri);
+  uris[0] = gtk_file_system_path_to_uri (path_bar->file_system, button_data->path);
+  uris[1] = NULL;
 
-  gtk_selection_data_set (selection_data,
-			  selection_data->target,
-			  8,
-			  (guchar *)uri_list,
-			  strlen (uri_list));
-  g_free (uri_list);
+  gtk_selection_data_set_uris (selection_data, uris);
+  g_free (uris[0]);
 }
 
 static ButtonData *
@@ -1281,10 +1442,7 @@ make_directory_button (GtkPathBar  *path_bar,
 		       gboolean     current_dir,
 		       gboolean     file_is_hidden)
 {
-  const GtkTargetEntry targets[] = {
-    { "text/uri-list", 0, 0 }
-  };
-
+  AtkObject *atk_obj;
   GtkWidget *child = NULL;
   GtkWidget *label_alignment = NULL;
   ButtonData *button_data;
@@ -1295,6 +1453,7 @@ make_directory_button (GtkPathBar  *path_bar,
 
   button_data->type = find_button_type (path_bar, path);
   button_data->button = gtk_toggle_button_new ();
+  atk_obj = gtk_widget_get_accessible (button_data->button);
   gtk_button_set_focus_on_click (GTK_BUTTON (button_data->button), FALSE);
 
   switch (button_data->type)
@@ -1303,6 +1462,7 @@ make_directory_button (GtkPathBar  *path_bar,
       button_data->image = gtk_image_new ();
       child = button_data->image;
       button_data->label = NULL;
+      atk_object_set_name (atk_obj, _("File System Root"));
       break;
     case HOME_BUTTON:
     case DESKTOP_BUTTON:
@@ -1348,9 +1508,9 @@ make_directory_button (GtkPathBar  *path_bar,
 
   gtk_drag_source_set (button_data->button,
 		       GDK_BUTTON1_MASK,
-		       targets,
-		       G_N_ELEMENTS (targets),
+		       NULL, 0,
 		       GDK_ACTION_COPY);
+  gtk_drag_source_add_uri_targets (button_data->button);
   g_signal_connect (button_data->button, "drag_data_get",
 		    G_CALLBACK (button_drag_data_get_cb), button_data);
 
@@ -1592,7 +1752,6 @@ _gtk_path_bar_set_file_system (GtkPathBar    *path_bar,
 			       GtkFileSystem *file_system)
 {
   const char *home;
-  char *desktop;
 
   g_return_if_fail (GTK_IS_PATH_BAR (path_bar));
 
@@ -1603,13 +1762,17 @@ _gtk_path_bar_set_file_system (GtkPathBar    *path_bar,
   home = g_get_home_dir ();
   if (home != NULL)
     {
+      const gchar *desktop;
+
       path_bar->home_path = gtk_file_system_filename_to_path (path_bar->file_system, home);
       /* FIXME: Need file system backend specific way of getting the
        * Desktop path.
        */
-      desktop = g_build_filename (home, "Desktop", NULL);
-      path_bar->desktop_path = gtk_file_system_filename_to_path (path_bar->file_system, desktop);
-      g_free (desktop);
+      desktop = g_get_user_special_dir (G_USER_DIRECTORY_DESKTOP);
+      if (desktop != NULL)
+        path_bar->desktop_path = gtk_file_system_filename_to_path (path_bar->file_system, desktop);
+      else 
+        path_bar->desktop_path = NULL;
     }
   else
     {
