@@ -84,7 +84,7 @@ enum {
 
 static guint socket_signals[LAST_SIGNAL] = { 0 };
 
-/**
+/*
  * _gtk_socket_get_private:
  *
  * @socket: a #GtkSocket
@@ -98,7 +98,7 @@ _gtk_socket_get_private (GtkSocket *socket)
   return G_TYPE_INSTANCE_GET_PRIVATE (socket, GTK_TYPE_SOCKET, GtkSocketPrivate);
 }
 
-G_DEFINE_TYPE (GtkSocket, gtk_socket, GTK_TYPE_CONTAINER);  
+G_DEFINE_TYPE (GtkSocket, gtk_socket, GTK_TYPE_CONTAINER)
 
 static void
 gtk_socket_finalize (GObject *object)
@@ -144,6 +144,13 @@ gtk_socket_class_init (GtkSocketClass *class)
   container_class->remove = gtk_socket_remove;
   container_class->forall = gtk_socket_forall;
 
+  /**
+   * GtkSocket::plug-added:
+   * @socket_: the object which received the signal
+   *
+   * This signal is emitted when a client is successfully
+   * added to the socket. 
+   */
   socket_signals[PLUG_ADDED] =
     g_signal_new (I_("plug_added"),
 		  G_OBJECT_CLASS_TYPE (class),
@@ -152,6 +159,17 @@ gtk_socket_class_init (GtkSocketClass *class)
 		  NULL, NULL,
 		  _gtk_marshal_VOID__VOID,
 		  G_TYPE_NONE, 0);
+
+  /**
+   * GtkSocket::plug-removed:
+   * @socket_: the object which received the signal
+   *
+   * This signal is emitted when a client is removed from the socket. 
+   * The default action is to destroy the #GtkSocket widget, so if you 
+   * want to reuse it you must add a signal handler that returns %TRUE. 
+   *
+   * Return value: %TRUE to stop other handlers from being invoked.
+   */
   socket_signals[PLUG_REMOVED] =
     g_signal_new (I_("plug_removed"),
 		  G_OBJECT_CLASS_TYPE (class),
@@ -285,13 +303,10 @@ gtk_socket_get_id (GtkSocket *socket)
 static void
 gtk_socket_realize (GtkWidget *widget)
 {
-  GtkSocket *socket;
+  GtkSocket *socket = GTK_SOCKET (widget);
   GdkWindowAttr attributes;
   gint attributes_mask;
 
-  g_return_if_fail (GTK_IS_SOCKET (widget));
-
-  socket = GTK_SOCKET (widget);
   GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
 
   attributes.window_type = GDK_WINDOW_CHILD;
@@ -319,8 +334,6 @@ gtk_socket_realize (GtkWidget *widget)
 			 _gtk_socket_windowing_filter_func,
 			 widget);
 
-  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
-
   /* We sync here so that we make sure that if the XID for
    * our window is passed to another application, SubstructureRedirectMask
    * will be set by the time the other app creates its window.
@@ -342,7 +355,7 @@ _gtk_socket_end_embedding (GtkSocket *socket)
   GtkWidget *toplevel = gtk_widget_get_toplevel (GTK_WIDGET (socket));
   gint i;
   
-  if (toplevel && GTK_IS_WINDOW (toplevel))
+  if (GTK_IS_WINDOW (toplevel))
     _gtk_socket_windowing_end_embedding_toplevel (socket);
 
   g_object_unref (socket->plug_window);
@@ -411,12 +424,7 @@ static void
 gtk_socket_size_allocate (GtkWidget     *widget,
 			  GtkAllocation *allocation)
 {
-  GtkSocket *socket;
-
-  g_return_if_fail (GTK_IS_SOCKET (widget));
-  g_return_if_fail (allocation != NULL);
-
-  socket = GTK_SOCKET (widget);
+  GtkSocket *socket = GTK_SOCKET (widget);
 
   widget->allocation = *allocation;
   if (GTK_WIDGET_REALIZED (widget))
@@ -731,11 +739,7 @@ static gboolean
 gtk_socket_focus (GtkWidget       *widget,
 		  GtkDirectionType direction)
 {
-  GtkSocket *socket;
-
-  g_return_val_if_fail (GTK_IS_SOCKET (widget), FALSE);
-  
-  socket = GTK_SOCKET (widget);
+  GtkSocket *socket = GTK_SOCKET (widget);
 
   if (socket->plug_widget)
     return gtk_widget_child_focus (socket->plug_widget, direction);
@@ -872,7 +876,7 @@ _gtk_socket_add_window (GtkSocket       *socket,
       /* Add a pointer to the socket on our toplevel window */
 
       toplevel = gtk_widget_get_toplevel (GTK_WIDGET (socket));
-      if (toplevel && GTK_IS_WINDOW (toplevel))
+      if (GTK_IS_WINDOW (toplevel))
 	gtk_window_add_embedded_xid (GTK_WINDOW (toplevel), xid);
 
       _gtk_socket_windowing_embed_notify (socket);
