@@ -69,19 +69,17 @@ static void gtk_toggle_button_update_state  (GtkButton            *button);
 
 static guint toggle_button_signals[LAST_SIGNAL] = { 0 };
 
-G_DEFINE_TYPE (GtkToggleButton, gtk_toggle_button, GTK_TYPE_BUTTON);
+G_DEFINE_TYPE (GtkToggleButton, gtk_toggle_button, GTK_TYPE_BUTTON)
 
 static void
 gtk_toggle_button_class_init (GtkToggleButtonClass *class)
 {
   GObjectClass *gobject_class;
   GtkWidgetClass *widget_class;
-  GtkContainerClass *container_class;
   GtkButtonClass *button_class;
 
   gobject_class = G_OBJECT_CLASS (class);
   widget_class = (GtkWidgetClass*) class;
-  container_class = (GtkContainerClass*) class;
   button_class = (GtkButtonClass*) class;
 
   gobject_class->set_property = gtk_toggle_button_set_property;
@@ -194,6 +192,7 @@ gtk_toggle_button_set_property (GObject      *object,
       gtk_toggle_button_set_mode (tb, g_value_get_boolean (value));
       break;
     default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
     }
 }
@@ -243,12 +242,8 @@ void
 gtk_toggle_button_set_mode (GtkToggleButton *toggle_button,
 			    gboolean         draw_indicator)
 {
-  GtkWidget *widget;
-
   g_return_if_fail (GTK_IS_TOGGLE_BUTTON (toggle_button));
 
-  widget = GTK_WIDGET (toggle_button);
-
   draw_indicator = draw_indicator ? TRUE : FALSE;
 
   if (toggle_button->draw_indicator != draw_indicator)
